#!   /usr/bin/env   python
#    coding: utf8

# python bindings for the etherbone unix library
# this is a preliminary version that gives access to the
# core C99 library API and implements Socket, Device and
# Cycle classes with a Pythonic interface
#
# just import-tested, probably full of horrendous bugs, needs
# extensive testing given that opaque types and nested
# structures are used all over the place and this is very
# error-prone with ctypes (and my limited knowledge thereof)

from ctypes import *
import logging

libetherbone_path = '/usr/local/lib/libetherbone.so'
lib = CDLL(libetherbone_path)

# typedefs and enums are translated into ctypes mainly
# for documentation purposes; using them does not make
# much sense otherwise

# configurable types
eb_address_t = c_ulonglong
eb_data_t = c_ulonglong

# control types
eb_network_address_t = c_char_p
eb_descriptor_t = c_int

# opaque structural types
# uncomment this and kill later definitions
# of eb_{socket,device,cycle}_t if some conflict
# with ctypes internals arises
eb_socket_t = c_uint16
eb_device_t = c_uint16
eb_cycle_t = c_uint16

# status codes
eb_err_code = {
        0:  'EB_OK',
        -1: 'EB_FAIL',
        -2: 'EB_ABORT',
        -3: 'EB_ADDRESS',
        -4: 'EB_OVERFLOW',
        -5: 'EB_ENDIAN',
        -6: 'EB_BUSY',
        -7: 'EB_TIMEOUT',
        -8: 'EB_OOM',
        -9: 'EB_ABI',
        -10: 'EB_SEGFAULT'
        }

# modes
eb_mode_t = c_int
EB_UNDEFINED = -1
EB_FIFO = 0
EB_LINEAR = 1

# bitmasks
eb_widths_t = c_uint
EB_UDP_MODE	= 	0
EB_FEC_MODE	= 	1

# widths
eb_width_t = c_uint
EB_ADDR8	= 	0x10
EB_ADDR16	= 	0x20
EB_ADDR32	= 	0x40
EB_ADDR64	= 	0x80
EB_ADDRX	= 	0xf0

EB_DATA8	= 	1
EB_DATA16	= 	2
EB_DATA32	= 	4
EB_DATA64	= 	8
EB_DATAX	= 	0xf

# etherbone.h
EB_ABI_CODE     =       0x488
EB_ENDIAN_MASK	=       0x30
EB_BIG_ENDIAN	=       0x10
EB_LITTLE_ENDIAN =      0x20

# callback types
eb_user_data_t = c_void_p

class Socket():
    """implement etherbone socket functionality (open, close, poll)

    All method return codes are identical to their plain C interface
    counterparts.
    """
    def __init__(self, port=0, widths=EB_ADDRX|EB_DATAX):
        self.abi_code = EB_ABI_CODE
        self.socket = eb_socket_t()
        self.port = port
        self.widths = widths

    def __enter__(self):
        status = self.open(self.port, self.widths, self.socket)
        if (status != 0):
            raise IOError('error opening device, code ', eb_err_code[status])
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def open(self, port, widths, socket):
        """open an Etherbone socket for communicating with remote devices

        The port parameter is optional; 0 lets the operating system choose.
        After opening the socket, poll must be hooked into an event loop.
        """
        return lib.eb_socket_open(self.abi_code, port, widths, byref(socket))

    def close(self):
        """close Etherbone socket

        Any use of the socket after successful close will probably segfault!
        """
        return lib.eb_socket_close(self.socket)

    def poll(self):
        """poll the Etherbone socket for activity
        """
        return lib.eb_socket_poll(self.socket)


class Device():
    """etherbone device interface (open, close, flush, read, write)

    """

    def __init__(self, socket, address):
        self.socket = socket.socket
        self.address = address
        self.widths = socket.widths
        self.device = eb_device_t()
        self.data_format = EB_DATA32|EB_BIG_ENDIAN

    def __enter__(self):
        status = self.open(self.socket, self.address, self.widths)
        if (status != 0):
            raise IOError('error opening device, code ', eb_err_code[status])
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def open(self, socket, ip_port, widths, retry=3):
        """open a remote Etherbone device

        This resolves the address and performs Etherbone end-point discovery.
        From the mask of proposed bus widths, one will be selected.
        The default port is taken as 0xEBD0.
        """
        return lib.eb_device_open(socket, ip_port, widths, retry, byref(self.device))

    def close(self):
        """close a remote Etherbone device
        """
        return lib.eb_device_close(self.device)

    def socket(self):
        """access the socket backing this device
        """
        return lib.eb_device_socket(self.device)

    def width(self):
        """recover the negotiated data width of the target device
        """
        return lib.eb_device_width(self.device)

    def read(self, address, user_data = None, callback = None):
        """perform a single-read wishbone cycle

        Semantically equivalent to cycle_open, cycle_read, cycle_close.

        The given address is read on the remote device.
        The callback cb(user, status, data) is invoked with the result.
        The user parameter is passed through uninspected to the callback.
        """
        data = eb_data_t()
        lib.eb_device_read(self.device, address, self.data_format, byref(data), user_data, callback)
        return data.value

    def write(self, address, data, user_data = None, callback = None):
        """perform a single-write wishbone cycle

        Semantically equivalent to cycle_open, cycle_write, cycle_close.
        data is written to the given address on the remote device.
        """
        return lib.eb_device_write(self.device, address, self.data_format, data, user_data, callback)

class Cycle():
    """provide wishbone cycle operations (open, close, read/write)

    Return codes for functions are identical to the plain C counterparts
    """

    def __init__(self, device, user_data=None, callback=None):
        self.dev_handle = device.device
        self.user_data = user_data
        self.callback = callback
        self.cycle = eb_cycle_t()
        self.data_format = device.data_format

    def __enter__(self):
        status = self.open(self.dev_handle, self.user_data, self.callback)
        if (status != 0):
            raise IOError('error opening device, code ', eb_err_code[status])
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def open(self, dev_handle, user_data=None, callback=None):
        return lib.eb_cycle_open(dev_handle, user_data, callback, byref(self.cycle))

    def close(self):
        """end a wishbone cycle

        This places the complete cycle at end of the device's send queue.
        You will probably want to flush() the underlying device()
        soon after calling this method
        """
        return lib.eb_cycle_close(self.cycle)

    def device(self):
        """access the device targetted by this cycle
        """
        return lib.eb_cycle_device(self.cycle)

    def read(self, address):
        """prepare a wishbone read phase

        The given address is read from the remote device.
        """
        data = eb_data_t()
        lib.eb_cycle_read(self.cycle, address, self.data_format, byref(data))
        return data.value

    def read_fast(self, address, size):
        """prepare a wishbone read phase

        The given address is read from the remote device.
        """
        dlist = []
        data = eb_data_t()
        for i in range(size):
            lib.eb_cycle_read(self.cycle, address, self.data_format, byref(data))
            logging.debug('read value 0x%x', data.value)
            dlist.append(data.value)
        return dlist

    def write(self, address, data):
        """perform a wishbone write phase

        data is written to the current cursor on the remote device.
        If the device was read-only, the operation is discarded.
        """
        logging.debug('write value 0x%x', data)
        d = eb_data_t(data)
        return lib.eb_cycle_write(self.cycle, address, self.data_format, d)
