library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.gn4124_core_pkg.all;
use work.gencores_pkg.all;
use work.wrcore_pkg.all;
use work.wr_fabric_pkg.all;
use work.wr_xilinx_pkg.all;
use work.wishbone_pkg.all;
use work.etherbone_pkg.all;

entity cute_top is
port
(
    -- Global ports
    clk_20m_vcxo_i : in std_logic;    -- 20MHz VCXO clock

    clk_125m_pllref_p_i : in std_logic;  -- 125 MHz PLL reference
    clk_125m_pllref_n_i : in std_logic;

    fpga_pll_ref_clk_101_p_i : in std_logic;  -- Dedicated clock for Xilinx GTP transceiver
    fpga_pll_ref_clk_101_n_i : in std_logic;

    -- Font panel LEDs
    led_red   : out std_logic;
    led_green : out std_logic;
    led_test  : out std_logic;    

    dac_sclk_o  : out std_logic;
    dac_din_o   : out std_logic;
    dac_clr_n_o : out std_logic;
    dac_ldac_n_o : out std_logic;
    dac_sync_n_o : out std_logic;
    fpga_scl_b : inout std_logic;
    fpga_sda_b : inout std_logic;
    thermo_id_b : inout std_logic;      -- 1-Wire interface to DS18B20

    -------------------------------------------------------------------------
    -- SFP pins
    -------------------------------------------------------------------------
    sfp_txp_o : out std_logic;
    sfp_txn_o : out std_logic;
    sfp_rxp_i : in std_logic;
    sfp_rxn_i : in std_logic;
    sfp_mod_def0_i    : in    std_logic;  -- sfp detect
    sfp_mod_def1_b    : inout std_logic;  -- scl
    sfp_mod_def2_b    : inout std_logic;  -- sda
    sfp_rate_select_b : inout std_logic;
    sfp_tx_fault_i    : in    std_logic;
    sfp_tx_disable_o  : out   std_logic;
    sfp_los_i         : in    std_logic;
    -----------------------------------------
    --UART
    -----------------------------------------
    uart_rxd_i : in  std_logic;
    uart_txd_o : out std_logic;
    ----------------------------------------
    -- PPS
    ---------------------------------------
    pps_o : out std_logic
);
end cute_top;

architecture rtl of cute_top is

------------------------------------------------------------------------------
-- Components declaration
------------------------------------------------------------------------------
component cute_wrc is
  generic
    (
      g_etherbone_enable: boolean:= true;
      g_multiboot_enable: boolean:= true
     );
  port
    (
      clk_20m_i     : in std_logic;
      clk_sys_i     : in std_logic;     -- 62.5m system clock, from pll drived by clk_125m_pllref
      clk_dmtd_i    : in std_logic;     -- 62.5m dmtd clock, from pll drived by clk_20m_vcxo
      clk_ref_i     : in std_logic;     -- 125m reference clock
      clk_gtp_i     : in std_logic;     -- dedicated clock for xilinx gtp transceiver

      rst_n_i     : in std_logic;

      -- font panel leds
      led_red   : out std_logic;
      led_green : out std_logic;
      led_test  : out std_logic;

      dac_sclk_o  : out std_logic;
      dac_din_o   : out std_logic;
      dac_clr_n_o : out std_logic;
      dac_ldac_n_o : out std_logic;
      dac_sync_n_o : out std_logic;

      fpga_scl_i : in  std_logic;
      fpga_scl_o : out std_logic;
      fpga_sda_i : in  std_logic;
      fpga_sda_o : out std_logic;

      --button1_i : in std_logic := 'h';
      --button2_i : in std_logic := 'h';

      ----------------------------------------
      -- Flash memory SPI
      ---------------------------------------
      fpga_prom_cclk_o       : out std_logic;
      fpga_prom_cso_b_n_o    : out std_logic;
      fpga_prom_mosi_o       : out std_logic;
      fpga_prom_miso_i       : in  std_logic:='1';

      thermo_id_i : in  std_logic;
      thermo_id_o : out std_logic;      -- 1-wire interface to ds18b20

      -------------------------------------------------------------------------
      -- sfp pins
      -------------------------------------------------------------------------
      sfp_txp_o : out std_logic;
      sfp_txn_o : out std_logic;
      sfp_rxp_i : in std_logic;
      sfp_rxn_i : in std_logic;
      sfp_mod_def0_i    : in    std_logic;  -- sfp detect
      sfp_mod_def1_i    : in std_logic;  -- scl
      sfp_mod_def1_o    : out std_logic;  -- scl
      sfp_mod_def2_i    : in std_logic;  -- sda
      sfp_mod_def2_o    : out std_logic;  -- sda
      sfp_rate_select_i : in std_logic;
      sfp_rate_select_o : out std_logic;
      sfp_tx_fault_i    : in    std_logic;
      sfp_tx_disable_o  : out   std_logic;
      sfp_los_i         : in    std_logic;

      pps_o : out std_logic;
      tm_time_valid_o      : out std_logic;
      tm_tai_o             : out std_logic_vector(39 downto 0);
      tm_cycles_o          : out std_logic_vector(27 downto 0);

      -----------------------------------------
      --uart
      -----------------------------------------
      uart_rxd_i : in  std_logic;
      uart_txd_o : out std_logic;

      ------------------------------------------
      -- external module
      ------------------------------------------
      ext_snk_i : in  t_wrf_sink_in;
      ext_snk_o : out t_wrf_sink_out;

      ext_src_o : out t_wrf_source_out;
      ext_src_i : in  t_wrf_source_in;

      ext_master_i : in t_wishbone_master_in:=cc_unused_master_in;
      ext_master_o : out t_wishbone_master_out
      );

end component;

component xwr_com5402 is
generic(
    g_use_wishbone_interface : boolean := true;
      -- use wishbone interface to configure ip/mac addr
    nudptx: integer range 0 to 1:= 1;
    nudprx: integer range 0 to 1:= 1;
      -- number of udp ports enabled for tx and rx
    ntcpstreams: integer range 0 to 255 := 1;  
      -- number of concurrent tcp streams handled by this component
    clk_frequency: integer := 120;
      -- clk frequency in mhz. needed to compute actual delays.
    tx_idle_timeout: integer range 0 to 50:= 50;  
      -- inactive input timeout, expressed in 4us units. -- 50*4us = 200us 
      -- controls the transmit stream segmentation: data in the elastic buffer will be transmitted if
      -- no input is received within tx_idle_timeout, without waiting for the transmit frame to be filled with mss data bytes.
    simulation: std_logic := '0'
      -- 1 during simulation with wireshark .cap file, '0' otherwise
      -- wireshark many not be able to collect offloaded checksum computations.
      -- when simulation =  '1': (a) ip header checksum is valid if 0000,
      -- (b) tcp checksum computation is forced to a valid 00001 irrespective of the 16-bit checksum
      -- captured by wireshark.
);
port(
    rst_n_i           : in std_logic;
    clk_ref_i         : in std_logic;
    clk_sys_i         : in std_logic;

    snk_i             : in  t_wrf_sink_in;
    snk_o             : out t_wrf_sink_out;
    src_o             : out t_wrf_source_out;
    src_i             : in  t_wrf_source_in;

    udp_rx_data       : out std_logic_vector(7 downto 0);
    udp_rx_data_valid : out std_logic;
    udp_rx_sof        : out std_logic;
    udp_rx_eof        : out std_logic;

    udp_tx_data       : in  std_logic_vector(7 downto 0):= (others=>'0');
    udp_tx_data_valid : in  std_logic:= '0';
    udp_tx_sof        : in  std_logic:= '0';
    udp_tx_eof        : in  std_logic:= '0';
    udp_tx_cts        : out std_logic;
    udp_tx_ack        : out std_logic;
    udp_tx_nak        : out std_logic;

    --// user-initiated connection reset for stream i
    connection_reset: in std_logic_vector((ntcpstreams-1) downto 0):=(others=>'0');

    tcp_rx_data       : out std_logic_vector(7 downto 0);
    tcp_rx_data_valid : out std_logic;
    tcp_rx_rts        : out std_logic;
    tcp_rx_cts        : in  std_logic:='1';
    tcp_tx_data       : in  std_logic_vector(7 downto 0):= (others=>'0');
    tcp_tx_data_valid : in  std_logic:='0';
    tcp_tx_cts        : out std_logic;

    cfg_slave_in  : in t_wishbone_slave_in:=cc_dummy_slave_in;
    cfg_slave_out : out t_wishbone_slave_out;

    tcp_local_port_no: in std_logic_vector(15 downto 0):=x"dcba";
    udp_rx_dest_port_no: in std_logic_vector(15 downto 0):=x"abcd";
    udp_tx_dest_ip_addr: in std_logic_vector(31 downto 0):=x"c0ab0201";
    udp_tx_source_port_no: in std_logic_vector(15 downto 0):=x"abcd";
    udp_tx_dest_port_no: in std_logic_vector(15 downto 0):=x"abcd"
);
end component;

component cute_reset_gen is
port (
    clk_sys_i : in std_logic;
    rst_n_o : out std_logic);
end component;

component user_tcp_demo is
port (
    clk_i : in std_logic;
    rst_n_i: in std_logic;

    tcp_rx_data: in std_logic_vector(7 downto 0);
    tcp_rx_data_valid:in std_logic;
    tcp_tx_data: out std_logic_vector(7 downto 0);
    tcp_tx_data_valid:out std_logic;
    tcp_tx_cts: in std_logic;
    tcp_rx_rts: in std_logic
	);
end component;

  ------------------------------------------------------------------------------
  -- Signals declaration
  ------------------------------------------------------------------------------
  -- Reset
  signal rst_n_i:std_logic;
  -- Clock
  signal clk_125m_pllref_i:std_logic;
  signal clk_ref_i,clk_sys_i: std_logic;
  signal clk_gtp_i,clk_dmtd_i:std_logic;
  signal clk_20m_vcxo_buf:std_logic;
  signal pllout_clk_62_5,pllout_clk_125:std_logic;
  signal pllout_clk_fb_ref,pllout_clk_fb_dmtd:std_logic;
  signal pllout_clk_dmtd:std_logic;
  
  signal dac_hpll_load_p1 : std_logic;
  signal dac_dpll_load_p1 : std_logic;
  signal dac_hpll_data    : std_logic_vector(15 downto 0);
  signal dac_dpll_data    : std_logic_vector(15 downto 0);

  signal pps     : std_logic;
  signal pps_led : std_logic;
  signal tm_utc  : std_logic_vector(39 downto 0);

  signal brd_temp:std_logic_vector(31 downto 0);
  signal temp:std_logic_vector(31 downto 0);
  signal brd_temp_valid:std_logic;
  
  signal sfp_mod_def1_i,sfp_mod_def1_o:std_logic;
  signal sfp_mod_def2_i,sfp_mod_def2_o:std_logic;
  signal fpga_scl_o : std_logic;
  signal fpga_scl_i : std_logic;
  signal fpga_sda_o : std_logic;
  signal fpga_sda_i : std_logic;
  signal thermo_id_i: std_logic;
  signal thermo_id_o: std_logic;
	
  signal ext_snk_i : t_wrf_sink_in;
  signal ext_snk_o : t_wrf_sink_out;
  signal ext_src_o : t_wrf_source_out:=c_dummy_snk_in;
  signal ext_src_i : t_wrf_source_in;  
  signal ext_slave_in:t_wishbone_slave_in;
	signal ext_slave_out:t_wishbone_slave_out;
	
  signal tcp_rx_data       : std_logic_vector(7 downto 0):= (others=>'0');
  signal tcp_rx_data_valid : std_logic;
  signal tcp_rx_rts        : std_logic;
  signal tcp_rx_cts        : std_logic;

  signal tcp_tx_data       : std_logic_vector(7 downto 0):= (others=>'0');
  signal tcp_tx_data_valid : std_logic;
  signal tcp_tx_cts        : std_logic;

  signal ext_my_mac_addr   :    std_logic_vector(47 downto 0):=x"2233076cb3b5"; -- debug mac address
  signal ext_my_ip_addr    :    std_logic_vector(31 downto 0):=x"c0a80024";
	signal ext_my_subnet_mask:    std_logic_vector(31 downto 0):=x"ffffff00";
	signal ext_my_gateway    :    std_logic_vector(31 downto 0):=x"c0a80001";
  signal ext_tcp_local_port_no: std_logic_vector(15 downto 0):=x"dcba";
  signal ext_udp_rx_dest_port_no:  std_logic_vector(15 downto 0):=x"dcba";
  signal ext_udp_tx_dest_ip_addr:  std_logic_vector(31 downto 0):=x"c0a80001";
  signal ext_udp_tx_source_port_no:  std_logic_vector(15 downto 0):=x"abcd";
  signal ext_udp_tx_dest_port_no:  std_logic_vector(15 downto 0):=x"abcd";
	
begin

U_Reset_Gen : cute_reset_gen
port map (
    clk_sys_i        => clk_sys_i,
    rst_n_o          => rst_n_i
);

cmp_refclk_buf : IBUFGDS
generic map (
    DIFF_TERM    => true,             -- Differential Termination
    IBUF_LOW_PWR => true,  -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
    IOSTANDARD   => "DEFAULT")
port map (
    O  => clk_125m_pllref_i,            -- Buffer output
    I  => clk_125m_pllref_p_i,  -- Diff_p buffer input (connect directly to top-level port)
    IB => clk_125m_pllref_n_i  -- Diff_n buffer input (connect directly to top-level port)
);

cmp_clk_vcxo_buf : BUFG
port map (
    O => clk_20m_vcxo_buf,
    I => clk_20m_vcxo_i
);

cmp_gtp_dedicated_clk_buf : IBUFGDS
generic map(
    DIFF_TERM    => true,
    IBUF_LOW_PWR => true,
    IOSTANDARD   => "DEFAULT")
port map (
    O  => clk_gtp_i,
    I  => fpga_pll_ref_clk_101_p_i,
    IB => fpga_pll_ref_clk_101_n_i
);

cmp_sys_clk_pll : PLL_BASE
generic map (
    BANDWIDTH          => "OPTIMIZED",
    CLK_FEEDBACK       => "CLKFBOUT",
    COMPENSATION       => "INTERNAL",
    DIVCLK_DIVIDE      => 1,
    CLKFBOUT_MULT      => 8,
    CLKFBOUT_PHASE     => 0.000,
    CLKOUT0_DIVIDE     => 16,        -- 62.5 MHz
    CLKOUT0_PHASE      => 0.000,
    CLKOUT0_DUTY_CYCLE => 0.500,
    CLKOUT1_DIVIDE     => 8,         -- 125 MHz
    CLKOUT1_PHASE      => 0.000,
    CLKOUT1_DUTY_CYCLE => 0.500,
    CLKOUT2_DIVIDE     => 4,         -- 250 MHz
    CLKOUT2_PHASE      => 0.000,
    CLKOUT2_DUTY_CYCLE => 0.500,
    CLKIN_PERIOD       => 8.0,
    REF_JITTER         => 0.016)
port map (
    CLKFBOUT => pllout_clk_fb_ref,
    CLKOUT0  => pllout_clk_62_5,
    CLKOUT1  => pllout_clk_125,
    CLKOUT2  => open,
    CLKOUT3  => open,
    CLKOUT4  => open,
    CLKOUT5  => open,
    LOCKED   => open,
    RST      => '0',
    CLKFBIN  => pllout_clk_fb_ref,
    CLKIN    => clk_125m_pllref_i
);

cmp_dmtd_clk_pll : PLL_BASE
generic map (
    BANDWIDTH          => "OPTIMIZED",
    CLK_FEEDBACK       => "CLKFBOUT",
    COMPENSATION       => "INTERNAL",
    DIVCLK_DIVIDE      => 1,
    CLKFBOUT_MULT      => 50,
    CLKFBOUT_PHASE     => 0.000,
    CLKOUT0_DIVIDE     => 16,         -- 62.5 MHz
    CLKOUT0_PHASE      => 0.000,
    CLKOUT0_DUTY_CYCLE => 0.500,
    CLKOUT1_DIVIDE     => 16,         -- 62.5 MHz
    CLKOUT1_PHASE      => 0.000,
    CLKOUT1_DUTY_CYCLE => 0.500,
    CLKOUT2_DIVIDE     => 8,
    CLKOUT2_PHASE      => 0.000,
    CLKOUT2_DUTY_CYCLE => 0.500,
    CLKIN_PERIOD       => 50.0,
    REF_JITTER         => 0.016)
port map (
    CLKFBOUT => pllout_clk_fb_dmtd,
    CLKOUT0  => pllout_clk_dmtd,
    CLKOUT1  => open,
    CLKOUT2  => open,
    CLKOUT3  => open,
    CLKOUT4  => open,
    CLKOUT5  => open,
    LOCKED   => open,
    RST      => '0',
    CLKFBIN  => pllout_clk_fb_dmtd,
    CLKIN    => clk_20m_vcxo_buf
);

cmp_clk_sys_buf : BUFG
port map (
    O => clk_sys_i,
    I => pllout_clk_62_5
);
   
cmd_clk_ref_buf: BUFG
port map(
    O => clk_ref_i,
    I => pllout_clk_125
);

cmp_clk_dmtd_buf : BUFG
  port map (
    O => clk_dmtd_i,
    I => pllout_clk_dmtd
);

  pps_o <= pps;
  
U_WR_CORE : cute_wrc
generic map(
  g_etherbone_enable=>true,
  g_multiboot_enable=>false
)
port map (
    clk_20m_i   => clk_20m_vcxo_buf,
    clk_sys_i   => clk_sys_i,
    clk_dmtd_i  => clk_dmtd_i,
    clk_ref_i   => clk_ref_i,
    clk_gtp_i   => clk_gtp_i,
    rst_n_i     => rst_n_i,
    led_red     => led_red,
    led_green   => led_green,
    led_test    => led_test,
    dac_sclk_o  => dac_sclk_o,
    dac_din_o   => dac_din_o,
    dac_clr_n_o => dac_clr_n_o,
    dac_ldac_n_o=> dac_ldac_n_o,
    dac_sync_n_o=> dac_sync_n_o,
    fpga_scl_o  => fpga_scl_o,
		fpga_scl_i  => fpga_scl_i,
    fpga_sda_o  => fpga_sda_o,
		fpga_sda_i  => fpga_sda_i,
    thermo_id_i => thermo_id_i,
		thermo_id_o => thermo_id_o,
    sfp_txp_o   => sfp_txp_o,
    sfp_txn_o   => sfp_txn_o,
    sfp_rxp_i   => sfp_rxp_i,
    sfp_rxn_i   => sfp_rxn_i,
    sfp_mod_def0_i=> sfp_mod_def0_i,
    sfp_mod_def1_i=> sfp_mod_def1_i,
    sfp_mod_def1_o=> sfp_mod_def1_o,
		sfp_mod_def2_i=> sfp_mod_def2_i,
    sfp_mod_def2_o=> sfp_mod_def2_o,
		sfp_rate_select_i=> '1',
    sfp_rate_select_o=> open,
    sfp_tx_fault_i=> sfp_tx_fault_i,
    sfp_tx_disable_o=> sfp_tx_disable_o,
    sfp_los_i   => sfp_los_i,
    uart_rxd_i  => uart_rxd_i,
    uart_txd_o  => uart_txd_o,
    tm_tai_o    => tm_utc,
    pps_o       => pps,
		ext_snk_o           => ext_src_i,
		ext_snk_i           => ext_src_o,
    ext_src_o           => ext_snk_i,
    ext_src_i           => ext_snk_o, 
    ext_master_i          => ext_slave_out,
    ext_master_o          => ext_slave_in
		);

  u_xwr_com5402: xwr_com5402
  port map(
    rst_n_i           => rst_n_i,
    clk_ref_i         => clk_ref_i,
    clk_sys_i         => clk_sys_i,

    snk_i             => ext_snk_i,
    snk_o             => ext_snk_o,
    src_o             => ext_src_o,
    src_i             => ext_src_i,

--    udp_rx_data       => udp_rx_data,
--    udp_rx_data_valid => udp_rx_data_valid,
--    udp_rx_sof        => udp_rx_sof,
--    udp_rx_eof        => udp_rx_eof,
--
--    udp_tx_data       => udp_tx_data,
--    udp_tx_data_valid => udp_tx_data_valid,
--    udp_tx_sof        => udp_tx_sof,
--    udp_tx_eof        => udp_tx_eof,
--    udp_tx_cts        => udp_tx_cts,
--    udp_tx_ack        => udp_tx_ack,
--    udp_tx_nak        => udp_tx_nak,
    connection_reset  => (others=>'0'),
    
    tcp_rx_data       => tcp_rx_data,
    tcp_rx_data_valid => tcp_rx_data_valid,  
    tcp_tx_data       => tcp_tx_data,
    tcp_tx_data_valid => tcp_tx_data_valid,
    tcp_tx_cts        => tcp_tx_cts,
    
    cfg_slave_in      => ext_slave_in,
    cfg_slave_out     => ext_slave_out
);

Inst_tcp_demo: user_tcp_demo
port map(
	clk_i          => clk_ref_i,
	rst_n_i        => rst_n_i,
	tcp_rx_data    => tcp_rx_data,
	tcp_rx_data_valid  => tcp_rx_data_valid,
	tcp_tx_data    => tcp_tx_data,
	tcp_tx_data_valid  => tcp_tx_data_valid,
	tcp_tx_cts     => tcp_tx_cts,
	tcp_rx_rts     => tcp_rx_rts
	);

process(clk_sys_i)
begin
if rising_edge(clk_sys_i) then
    if (brd_temp_valid='1') then
         temp <= brd_temp;
    else
         temp <= (others=>'0');
    end if;
end if;
end process;

  fpga_scl_b  <= '0' when fpga_scl_o = '0' else 'Z';
  fpga_sda_b  <= '0' when fpga_sda_o = '0' else 'Z';
  fpga_scl_i  <= fpga_scl_b;
  fpga_sda_i  <= fpga_sda_b;

  sfp_mod_def1_b <= '0' when sfp_mod_def1_o = '0' else 'Z';
  sfp_mod_def2_b <= '0' when sfp_mod_def2_o = '0' else 'Z';
  sfp_mod_def1_i <= sfp_mod_def1_b;
  sfp_mod_def2_i <= sfp_mod_def2_b;

  thermo_id_b <= '0' when thermo_id_o = '1' else 'Z';
  thermo_id_i  <= thermo_id_b;
  
end rtl;


